""" invo - An Inverse Optimization Library

invo is a Python package intended to simplify the use of inverse optimization as
a model fitting practice. Our goal is to provide a general framework for 
formulating and solving inverse optimization problems, as well as implement a
collection of inverse methods. As additional algorithms and methods arise in 
the literature, we can implement them using the same common framework.

The flavour of this work is inspired from scikit-learn on Python.
"""



__all__ = ['LinearModels', 'utils']

